;################################################################################
;#										#
;# fsys - Dataflash library							#
;# copyright (c) 2007-2011 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;# This library is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU  General Public				#
;# License as published by the Free Software Foundation; either			#
;# version 3 of the License, or (at your option) any later version.		#
;#										#
;# This library is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU  General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

; 64 Pages / File

;fileio_ram+32 = tape number

fsys_init:
		sbi	SPI_SPORT,SPI_SPIN
		sbi	SPI_SDDR,SPI_SPIN
		sbi	MOSI_DDR,MOSI_PIN
		sbi	SCK_DDR,SCK_PIN
		ldi	XL,0x5c
		out	SPCR,XL
		sts	fileio_ram+32,const_0		;we start with
		sts	fileio_ram+36,const_0		;no tape
		rcall	fsys_stat		;wait for ready
		brtc	fsys_init_1
		sts	fileio_ram+36,const_1		;tape available
fsys_init_1:	ret

;------------------------------------------------------------------------------
; set enable
;------------------------------------------------------------------------------
fsys_sel:	cbi	PORTB,4			;now set SS signal
		ret

fsys_unsel:	sbi	PORTB,4
		ret

;------------------------------------------------------------------------------
; send dataflash byte (r17)
;------------------------------------------------------------------------------
fsys_outzero:	clr	r17
fsys_out:	out	SPDR,r17
fsys_out1:	in	r17,SPSR
		sbrs	r17,SPIF
		rjmp	fsys_out1
		in	r17,SPDR
		ret

;output dummy bytes
fsys_outd4:	rcall	fsys_outzero
fsys_outd3:	rcall	fsys_outzero
fsys_outd2:	rcall	fsys_outzero
		rjmp	fsys_outzero		;get byte


fsys_fpage0:	sts	fileio_ram+33,r16
fsys_fpage:	lds	r0,fileio_ram+32	;tape number
		lds	r16,fileio_ram+33	;file number
		andi	r16,0x1f		;limit to 32
		sbrc	r0,0
		ori	r16,0x20		;second tape
		ldi	XL,128			;pages per program*2
		mul	XL,r16
		movw	XL,r0			;this is now the first page
		andi	r16,0x1f		;limit to 32
		ret

;------------------------------------------------------------------------------
; wait for Dataflash ready status and write size
; get size to r17 0x08/0x10
; T is 1 if OK
;------------------------------------------------------------------------------
fsys_stat:	rcall	fsys_sel
		ldi	r17,0xd7		;command
		rcall	fsys_out		;output command
		rcall	fsys_outzero
		rcall	fsys_unsel
		andi	r17,0xbc		;mask out compare an x bits
		cpi	r17,0x24		;8Mb busy
		breq	fsys_stat
		cpi	r17,0x1c		;4Mb busy
		breq	fsys_stat
		;no busy state
		cpi	r17,0xa4		;8Mb OK
		breq	fsys_stat1
		cpi	r17,0x9c		;4Mb OK
		breq	fsys_stat1
fsys_error1:	clt				;error->no DF
		clr	r17			;no pages
		ret
fsys_stat1:	subi	r17,0x94		;calc pages
		set				;OK
		ret				;thats all

;------------------------------------------------------------------------------
; read block X to buffer 1
;------------------------------------------------------------------------------
fsys_bread1:	rcall	fsys_stat		;wait for ready
		rcall	fsys_sel
		ldi	r17,0x53
fsys_breadx:	rcall	fsys_out
		mov	r17,XH
		rcall	fsys_out		;out high
		mov	r17,XL
		rcall	fsys_out		;out low
		rcall	fsys_outzero		;00=byte in page
		rcall	fsys_unsel		;stop transfer
		rjmp	fsys_stat		;wait for ready

;------------------------------------------------------------------------------
; write block X from buffer 1
;------------------------------------------------------------------------------
fsys_bwrite1:	rcall	fsys_stat		;wait for ready
		rcall	fsys_sel
		ldi	r17,0x83
		rjmp	fsys_breadx


;------------------------------------------------------------------------------
; read DATA r18 data bytes of buf 1 to (Y+)
;------------------------------------------------------------------------------
fsys_rd1:	rcall	fsys_stat		;wait for ready
		rcall	fsys_sel
		ldi	r17,0xd4		;Buffer 1 Read
		rcall	fsys_out		;output command
		rcall	fsys_outd4		;ADDR + dummy byte
fsys_rd1_1:	rcall	fsys_outzero		;read byte
		st	Y+,r17
		dec	r18
		brne	fsys_rd1_1
		rjmp	fsys_unsel		;stop transfer

;------------------------------------------------------------------------------
; set buffer 1 to 0xff
;------------------------------------------------------------------------------
fsys_cl1:	rcall	fsys_stat		;wait for ready
		rcall	fsys_sel
		ldi	r17,0x84		;Buffer 1 Write
fsys_cl1_0:	rcall	fsys_out		;output command
		rcall	fsys_outd3		;ADDR
		clr	r0
fsys_cl1_1:	ldi	r17,0xff
		rcall	fsys_out		;write data
		dec	r0
		brne	fsys_cl1_1
		rjmp	fsys_unsel		;stop transfer

;------------------------------------------------------------------------------
; write DATA r18 data bytes (Y+) to buf 1
;------------------------------------------------------------------------------
fsys_wd1:	rcall	fsys_stat		;wait for ready
		rcall	fsys_sel
		ldi	r17,0x84		;Buffer 1 Write
		rcall	fsys_out		;output command
		rcall	fsys_outd3		;ADDR
fsys_wd1_1:	ld	r17,Y+
		rcall	fsys_out		;write data
		dec	r18
		brne	fsys_wd1_1
fsys_wd1_3:	rjmp	fsys_unsel		;stop transfer


;------------------------------------------------------------------------------
; read name of file r16 to (Y+)
;------------------------------------------------------------------------------
fsys_rname:	rcall	fsys_stat
		cpi	r17,0
		brne	fsys_rname_5
		ldi	r18,10
		ldi	r16,0x16		;"-"
		ldi	YL,LOW(fileio_ram+16)
		ldi	YH,HIGH(fileio_ram+16)
fsys_rname_1:	st	Y+,r16
		dec	r18
		brne	fsys_rname_1
		ret

fsys_rname_5:	rcall	fsys_fpage0
;		clr	XL
;		clr	XH
		rcall	fsys_bread1
		ldi	YL,LOW(fileio_ram+16)
		ldi	YH,HIGH(fileio_ram+16)
		ldi	r18,12			;read 12 bytes
		rjmp	fsys_rd1

;------------------------------------------------------------------------------
; find file
;------------------------------------------------------------------------------
fsys_ffile:	ldi	r16,0x1f		;we start with file 31
		ldi	ZL,0xff			;file number
		ldi	ZH,0xff			;free file
fsys_ffile_1:	push	r16
		rcall	fsys_rname		;get name
		pop	r16
		ldi	YL,LOW(fileio_ram+16)	;filename from file
		ldi	YH,HIGH(fileio_ram+16)
		ldi	XL,LOW(fileio_ram)	;filename from request
		ldi	XH,HIGH(fileio_ram)

		ldi	r19,10			;chars to compare
fsys_ffile_2:	ld	r17,Y+			;from file
		cpi	r17,0xff		;empty
		brne	fsys_ffile_3		;branch if not
		mov	ZH,r16			;set free file
		rjmp	fsys_ffile_4

fsys_ffile_3:	ld	r18,X+			;from req
		cp	r17,r18
		brne	fsys_ffile_4
		dec	r19
		brne	fsys_ffile_2
		mov	ZL,r16
		ret

fsys_ffile_4:	sub	r16,const_1
		brcc	fsys_ffile_1
		ret


;------------------------------------------------------------------------------
; write file
;------------------------------------------------------------------------------
fsys_wfile:	rcall	fsys_ffile		;find file
		cpi	ZL,0xff
		breq	fsys_wfile_1		;not found
		mov	r16,ZL			;this is our file to write
		rjmp	fsys_wfile_3		;start write
fsys_wfile_1:	cpi	ZH,0xff
		brne	fsys_wfile_2		;empty found
		rjmp	fsys_file_err

fsys_wfile_2:	mov	r16,ZH			;write o this empty file
		;now write
fsys_wfile_3:	rcall	fsys_fpage0
		rcall	fsys_cl1		;clear page
		ldi	YL,LOW(fileio_ram)	;filename from request
		ldi	YH,HIGH(fileio_ram)
		ldi	r18,12
		rcall	fsys_wd1		;write data to buffer
		rcall	fsys_bwrite1		;write buffer to flash array

		sbiw	YL,2
		ld	ZL,Y+			;file length
		ld	ZH,Y+
		ldi	YL,0x09			;start in RAM
		ldi	YH,0x02

		adiw	XL,2			;next page

fsys_wfile_4:	cpi	ZH,0			;full page
		breq	fsys_wfile_5		;branch if not
		ldi	r18,0			;full page
		rcall	fsys_wd1		;write data to buffer
		rcall	fsys_bwrite1		;write buffer to flash array
		adiw	XL,2
		dec	ZH
		rjmp	fsys_wfile_4

fsys_wfile_5:	cpi	ZL,0
		breq	fsys_wfile_6
		mov	r18,ZL			;full page
		rcall	fsys_wd1		;write data to buffer
		rcall	fsys_bwrite1		;write buffer to flash array

fsys_wfile_6:	set				;OK
		ret


fsys_file_err:	ldi	XL,T_ERR
		sts	0x200,XL
		ret

;------------------------------------------------------------------------------
; read file
;------------------------------------------------------------------------------
fsys_rfile:	rcall	fsys_ffile		;find file
		cpi	ZL,0xff
		breq	fsys_rfile_1		;not found
		mov	r16,ZL			;this is our file to write
		rjmp	fsys_rfile_02		;start write
fsys_rfile_1:	set
		ret				;file not found

fsys_rfile_02:	rcall	fsys_fpage0
		ldi	YL,LOW(fileio_ram)	;filename from request
		ldi	YH,HIGH(fileio_ram)
		rcall	fsys_bread1		;read buffer 1 from flash array
		ldi	r18,12
		rcall	fsys_rd1		;write data to buffer
		sbiw	YL,2
		ld	ZL,Y+			;file length
		ld	ZH,Y+
		ldi	YL,0x09			;start in RAM
		ldi	YH,0x02	

		adiw	XL,2			;next page

fsys_rfile_4:	cpi	ZH,0			;full page
		breq	fsys_rfile_5		;branch if not
		rcall	fsys_bread1		;read buffer 1 from flash array
		ldi	r18,0			;full page
		rcall	fsys_rd1		;write data to buffer
		adiw	XL,2
		dec	ZH
		rjmp	fsys_rfile_4

fsys_rfile_5:	cpi	ZL,0
		breq	fsys_rfile_6
		rcall	fsys_bread1		;read buffer 1 from flash array
		mov	r18,ZL			;full page
		rcall	fsys_rd1		;write data to buffer

fsys_rfile_6:	set				;OK
		ret

;------------------------------------------------------------------------------
; delete file
;------------------------------------------------------------------------------
fsys_dfile:	lds	r17,fileio_ram+33	;file number
		lds	r16,fileio_ram+32	;tape
		andi	r17,0x1f
		sbrc	r16,0
		ori	r17,0x20		;second tape
		ldi	XL,64			;pages per program
		mul	XL,r17
		movw	XL,r0			;this is now the first page
		rcall	fsys_cl1		;clear buffer
		rcall	fsys_bwrite1		;write buffer to flash array
		ret

.if USE_BL == 1
;read block
fsys_rblock:	rcall	fsys_fpage
		lds	r16,fileio_ram+35	;page in file
		add	XL,r16
		adc	XH,const_0
		add	XL,r16
		adc	XH,const_0
		rcall	fsys_bread1		;read page to block
		clr	r18
		ldi	YL,0
		ldi	YH,bl_buffer
		rcall	fsys_rd1
		adiw	XL,2
		rcall	fsys_bread1		;read page to block
		clr	r18
		rjmp	fsys_rd1

fsys_wblock:	rcall	fsys_fpage		;calc X=page
		lds	r16,fileio_ram+35	;page in file
		add	XL,r16			;add 2x
		adc	XH,const_0
		add	XL,r16
		adc	XH,const_0
		ldi	YL,0
		ldi	YH,bl_buffer
		clr	r18
		rcall	fsys_wd1
		rcall	fsys_bwrite1
		adiw	XL,2
		clr	r18
		rcall	fsys_wd1
		rjmp	fsys_bwrite1
.endif