;################################################################################
;#										#
;# fsys - SD-Card library							#
;# copyright (c) 2009-2011 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;# This library is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU  General Public				#
;# License as published by the Free Software Foundation; either			#
;# version 3 of the License, or (at your option) any later version.		#
;#										#
;# This library is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU  General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################


;------------------------------------------------------------------------------
; wait for SPI is ready for next data
; write XL to SPI
;------------------------------------------------------------------------------
fsys_write_ff2:	rcall	fsys_write_ff
fsys_write_ff:	ldi	XL,0xff			;dummy data
fsys_wspi:	out	SPDR,XL
fsys_wait:	in	XL,SPSR
		sbrs	XL,SPIF
		rjmp	fsys_wait
		in	XL,SPDR
		ret

;------------------------------------------------------------------------------
; write command XH to SPI
; XL=return byte (FF=time out)
;------------------------------------------------------------------------------
fsys_cmd:	rcall	fsys_write_ff		;8 dummy clocks
		cbi	SPI_SPORT,SPI_SPIN	;CS active
		out	SPDR,XH			;CMD
		rcall	fsys_wait
		out	SPDR,r19		;address MSB
		rcall	fsys_wait
		out	SPDR,r18
		rcall	fsys_wait
		out	SPDR,r17
		rcall	fsys_wait
		out	SPDR,r16		;address LSB
		rcall	fsys_wait
		ldi	XL,0x95			;dummy checksum
		rcall	fsys_wspi
		clr	r0
		clr	r1

fsys_cmd_resp:	rcall	fsys_write_ff		;write 0xff and read data
		cpi	XL,0xff
		brne	fsys_cmd_end
		add	r0,const_1
		adc	r1,const_0
		sbrs	r1,6			;16384 loops max
		rjmp	fsys_cmd_resp
fsys_cmd_end:	ret

;------------------------------------------------------------------------------
; wait for start block mtoken
;------------------------------------------------------------------------------
fsys_waitsbt:	push	ZL
		push	ZH
		ldi	ZL,0			;timeout
		ldi	ZH,0

fsys_waitsbt1:	cpi	XL,0xfe			;start block token
		brne	fsys_waitsbt2
		pop	ZH
		pop	ZL
		ret				;SBT found

fsys_waitsbt2:	sbiw	ZL,1			;timeout counter
		breq	fsys_waitsbt3		;branch if timeout is reached
		rcall	fsys_write_ff
		rjmp	fsys_waitsbt1		;next try

fsys_waitsbt3:	pop	ZH
		pop	ZL
		pop	r0			;kill last stack entry
		pop	r0
		rjmp	fsys_dfile_e

;------------------------------------------------------------------------------
; init SD card and determine size
;------------------------------------------------------------------------------
fsys_init:	out	SPCR,const_0
		sbi	SPI_SPORT,SPI_SPIN
		sbi	SPI_SDDR,SPI_SPIN
		sbi	MOSI_DDR,MOSI_PIN
		sbi	SCK_DDR,SCK_PIN
		sts	fileio_ram+32,const_0	;start with tape 0

		ldi	XL,0x53			;f/128
		out	SPCR,XL
		ldi	XH,16
fsys_init_01:	rcall	fsys_write_ff		;128 clock pulses
		dec	XH
		brne	fsys_init_01
		clr	r16			;ADDRFIELD=0
		clr	r17
		movw	r18,r16

		clr	YL			;CMD0 retries
fsys_init_02:	ldi	XH,0x40			;CMD0
		rcall	fsys_cmd		;send CMD0
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		cpi	XL,0x01			;idle state
		breq	fsys_init_03		;yes->OK
		dec	YL			;retry counter
		brne	fsys_init_02		;retry CMD 0

fsys_init_err:	sts	fileio_ram+36,const_0	;not available
		ret				;exit

fsys_init_03:	ldi	YL,0			;ACM41 retries
		ldi	YH,4

fsys_init_04:	ldi	XH,0x77			;CMD55
		rcall	fsys_cmd		;send CMD55
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		ldi	XH,0x69			;ACMD41
		rcall	fsys_cmd		;send ;ACMD41
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		cpi	XL,0x00			;OK?
		breq	fsys_init_end
		sbiw	YL,1			;timeout
		brne	fsys_init_04		;do timeout loop
		rjmp	fsys_init_err		;error

fsys_init_end:	clr	r16			;ADDRFIELD=0
		ldi	r17,2
		clr	r18
		clr	r19
		ldi	XH,0x50			;CMD16
		rcall	fsys_cmd		;send CMD16
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		cpi	XL,0x00			;idle state
		brne	fsys_init_end		;yes->OK
		ldi	XL,0x50			;f/4
		out	SPCR,XL
		sts	fileio_ram+36,const_1	;available
		ret

;------------------------------------------------------------------------------
; calc sector address to r16..r19
; fileio_ram+32	= tape
; fileio_ram+33	= file
; fileio_ram+35	= sector
;------------------------------------------------------------------------------
fsys_sectaddr0:	sts	fileio_ram+35,const_0	;first sector
fsys_sectaddr:	lds	XL,fileio_ram+32	;tape
		ldi	XH,8			;1024 sectors/tape (/128)
		mul	XL,XH
		movw	r18,r0
		lds	XL,fileio_ram+33	;file in tape
		andi	XL,0x1f
		ldi	XH,64			;64 * 256 bytes /file
		mul	XL,XH
		mov	r17,r0
		add	r18,r1
		adc	r19,const_0
		lds	r16,fileio_ram+35	;sector
		lsl	r16			;*2
		add	r17,r16
		adc	r18,const_0
		adc	r19,const_0
		clr	r16
		ret


;------------------------------------------------------------------------------
; read name
; fileio_ram+32	= tape
; fileio_ram+33	= file
;------------------------------------------------------------------------------
fsys_rname:	ldi	YL,LOW(fileio_ram+16)
		ldi	YH,HIGH(fileio_ram+16)

		lds	XL,fileio_ram+36
		sbrs	XL,0
		rjmp	fsys_dname		;dummy name

		rcall	fsys_sectaddr0		;calc sector addr
		ldi	XH,0x51			;read block
		rcall	fsys_cmd		;send CMD
		rcall	fsys_waitsbt		;wait for start block token
		ldi	XH,12			;fname length + len
fsys_rname_1:	rcall	fsys_write_ff
		st	Y+,XL
		dec	XH
		brne	fsys_rname_1
		ldi	XH,251			;(510/2)+2CRC
fsys_rname_2:	rcall	fsys_write_ff2
		dec	XH
		brne	fsys_rname_2
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		ret

fsys_dname:	ldi	XH,10
		ldi	XL,0x16
fsys_dname_1:	st	Y+,XL
		dec	XH
		brne	fsys_dname_1
		ret

;------------------------------------------------------------------------------
; read file
; fileio_ram+32	= tape
; fileio_ram+33	= file
;------------------------------------------------------------------------------
fsys_rfile:	rcall	fsys_ffile		;find file
		cpi	ZL,0xff			;is "not found" code?
		brne	fsys_rfile_03		;not found
fsys_rfile_01:	ldi	XL,T_ERR-1
		sts	0x200,XL		;error
		ret				;file not found

fsys_rfile_02:

fsys_rfile_03:	sts	fileio_ram+33,ZL	;file number
		rcall	fsys_sectaddr0		;calc sector addr
		ldi	YL,LOW(fileio_ram+16)
		ldi	YH,HIGH(fileio_ram+16)
		ldi	XH,0x51			;read block
		rcall	fsys_cmd		;send CMD
		rcall	fsys_waitsbt		;wait for start block token
		ldi	XH,12			;fname length + len
fsys_rfile_1:	rcall	fsys_write_ff
		st	Y+,XL
		dec	XH
		brne	fsys_rfile_1

		ldi	XH,244			;rest of 256 bytes
fsys_rfile_1a:	rcall	fsys_write_ff
		dec	XH
		brne	fsys_rfile_1a

		lds	ZL,(fileio_ram+26)	;LEN LOW
		lds	ZH,(fileio_ram+27)	;LEN HIGH
		ldi	YL,0x09			;start at 0x4009 in Z80 adr space
		ldi	YH,0x02
		mov	r1,const_1
		clr	r0

fsys_rfile_2:	rcall	fsys_write_ff		;get data byte
		cp	ZL,const_0
		cpc	ZH,const_0
		breq	fsys_rfile_3		;we must do a dummy read
		st	Y+,XL			;store data byte in RAM
		sbiw	ZL,1			;to-do bytes -1
fsys_rfile_3:	sub	r0,const_1
		sbc	r1,const_0
		brne	fsys_rfile_2
		rcall	fsys_write_ff2		;get CRC bytes
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive

		cp	ZL,const_0		;all done?
		cpc	ZH,const_0
		breq	fsys_rfile_e		;yes, we can exit

		lds	XL,fileio_ram+35	;next sector
		inc	XL
		sts	fileio_ram+35,XL
		rcall	fsys_sectaddr		;calc sector addr
		ldi	XH,0x51			;read block
		rcall	fsys_cmd		;send CMD
		rcall	fsys_waitsbt		;wait for start block token
		clr	r0
		ldi	XL,2
		mov	r1,XL			;512 bytes/sector

		rjmp	fsys_rfile_2

fsys_rfile_e:	ret				;all done


;------------------------------------------------------------------------------
; write file
; fileio_ram+32	= tape
; fileio_ram+33	= file
;------------------------------------------------------------------------------
fsys_wfile:	rcall	fsys_ffile		;find file
		cpi	ZL,0xff
		breq	fsys_wfile_01		;not found
		mov	r16,ZL			;this is our file to write
		rjmp	fsys_wfile_03		;start write
fsys_wfile_01:	cpi	ZH,0xff
		brne	fsys_wfile_02		;empty found
		rjmp	fsys_dfile_e

fsys_wfile_02:	mov	r16,ZH			;write to this empty file

fsys_wfile_03:	sts	fileio_ram+33,r16	;store file number

fsys_wfile_0:	rcall	fsys_sectaddr0		;calc sector addr
		ldi	XH,0x58			;write block
		rcall	fsys_cmd		;send CMD
		ldi	XL,0xfe			;start block token
		rcall	fsys_wspi		;send
		ldi	YL,LOW(fileio_ram)
		ldi	YH,HIGH(fileio_ram)
		ldi	XH,12			;fname length + len
fsys_wfile_1:	ld	XL,Y+
		rcall	fsys_wspi		;send
		dec	XH
		brne	fsys_wfile_1

		ldi	XH,244			;rest of 256 bytes
fsys_wfile_1a:	rcall	fsys_write_ff		;send
		dec	XH
		brne	fsys_wfile_1a

		lds	ZL,(fileio_ram+10)	;LEN LOW
		lds	ZH,(fileio_ram+11)	;LEN HIGH
		ldi	YL,0x09			;start at 0x4009 in Z80 adr space
		ldi	YH,0x02
		mov	r1,const_1
		clr	r0

fsys_wfile_2:	ldi	XL,0xff
		cp	ZL,const_0
		cpc	ZH,const_0
		ldi	XL,0xff
		breq	fsys_wfile_3		;we must do a dummy byte write
		ld	XL,Y+			;store data byte in RAM
		sbiw	ZL,1

fsys_wfile_3:	rcall	fsys_wspi		;send
		sub	r0,const_1
		sbc	r1,const_0
		brne	fsys_wfile_2
		rcall	fsys_write_ff2		;put dummy CRC bytes
fsys_wfile_4:	rcall	fsys_write_ff		;put dummy byte
		cpi	XL,0xff
		brne	fsys_wfile_4
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive

		cp	ZL,const_0		;all done?
		cpc	ZH,const_0
		breq	fsys_rfile_e		;yes, we can exit

		lds	XL,fileio_ram+35	;next sector
		inc	XL
		sts	fileio_ram+35,XL
		rcall	fsys_sectaddr

		ldi	XH,0x58			;write block
		rcall	fsys_cmd		;send CMD
		ldi	XL,0xfe			;start block token
		rcall	fsys_wspi		;send

		clr	r0
		ldi	XL,2
		mov	r1,XL			;512 bytes/sector

		rjmp	fsys_wfile_2

;------------------------------------------------------------------------------
; find file
;------------------------------------------------------------------------------
fsys_ffile:	ldi	r16,0x1f		;we start with file 31
		ldi	ZL,0xff			;file number
		ldi	ZH,0xff			;free file
fsys_ffile_1:	sts	fileio_ram+33,r16	;store file number
		rcall	fsys_rname		;get name
		ldi	YL,LOW(fileio_ram+16)	;filename from file
		ldi	YH,HIGH(fileio_ram+16)
		ldi	XL,LOW(fileio_ram)	;filename from request
		ldi	XH,HIGH(fileio_ram)

		ldi	r19,10			;chars to compare
fsys_ffile_2:	ld	r17,Y+			;from file
		cpi	r17,0xff		;empty
		brne	fsys_ffile_3		;branch if not
		lds	ZH,fileio_ram+33	;set free file

fsys_ffile_3:	ld	r18,X+			;from req
		cp	r17,r18
		brne	fsys_ffile_4
		dec	r19
		brne	fsys_ffile_2
		lds	ZL,fileio_ram+33	;found
		ret

fsys_ffile_4:	lds	r16,fileio_ram+33
		sub	r16,const_1
		brcc	fsys_ffile_1
		ret


; delete file
fsys_dfile:	rcall	fsys_sectaddr0		;calc sector addr
		ldi	XH,0x58			;write block
		rcall	fsys_cmd		;send CMD
		ldi	XL,0xfe			;start block token
		rcall	fsys_wspi		;send
		ldi	XH,0x00			;complete sector
fsys_dfile_1:	rcall	fsys_write_ff2		;send
		dec	XH
		brne	fsys_dfile_1
		rcall	fsys_write_ff2		;put dummy CRC bytes
fsys_dfile_4:	rcall	fsys_write_ff		;put dummy byte
		cpi	XL,0xff
		brne	fsys_dfile_4
		sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		ret

fsys_dfile_e:	ldi	XL,T_ERR
		sts	0x200,XL
		ret

.if USE_BL == 1
;read one block (for bootloader)
fsys_rblock:	rcall	fsys_sectaddr		;calc sector addr
		ldi	YL,0
		ldi	YH,bl_buffer
		ldi	XH,0x51			;read block
		rcall	fsys_cmd		;send CMD
		rcall	fsys_waitsbt		;wait for start block token
		ldi	XH,0			;(512)
fsys_rblock_1:	rcall	fsys_write_ff
		st	Y+,XL
		rcall	fsys_write_ff
		st	Y+,XL
		dec	XH
		brne	fsys_rblock_1
		rcall	fsys_write_ff2		;dummy read CRC
fsys_rblock_e:	sbi	SPI_SPORT,SPI_SPIN	;CS inactive
		ret

;write one block (for bootloader)
fsys_wblock:	rcall	fsys_sectaddr		;calc sector addr
		ldi	XH,0x58			;write block
		rcall	fsys_cmd		;send CMD
		ldi	XL,0xfe			;start block token
		rcall	fsys_wspi		;send
		ldi	YL,0
		ldi	YH,bl_buffer
		ldi	XH,0			;fname length + len
fsys_wblock_1:	ld	XL,Y+
		rcall	fsys_wspi		;send
		ld	XL,Y+
		rcall	fsys_wspi		;send
		dec	XH
		brne	fsys_wblock_1
		rcall	fsys_write_ff2		;put dummy CRC bytes
fsys_wblock_2:	rcall	fsys_write_ff		;put dummy byte
		cpi	XL,0xff
		brne	fsys_wblock_2
		rjmp	fsys_rblock_e
.endif
